@@ -6,4 +6,8 @@ from .text_diversities.phonological import *
 from .text_diversities.syntactic import *
 from .text_diversities.semantic import *
 
+from . import (
+    search
+)
+
 name = "textdiversity"
\ No newline at end of file
